import io
import csv
import shutil
import tempfile
import unittest
from pathlib import Path
from unittest import mock

from pxr import Usd, UsdGeom, Sdf, UsdShade

from grill import usd

try:
    from grill import cook, names
    _COOK_AVAILABLE = True
except ImportError as exc:
    _COOK_AVAILABLE = False
    print(f"'grill.cook' module failed to import. Unable to test: {exc.msg}")

from grill.views import description, sheets, create, _attributes, stats, _core, _graph, _qt
from grill.views._qt import QtWidgets, QtCore, QtGui

# 2024-02-03 - Python-3.12 & USD-23.11
# leaving the PySide6 import below freezes windows in Python-3.12. Importing it first when running tests "fixes" the freeze.
# from PySide6 import QtWebEngineWidgets
# alternatively, the following can be called to ensure shared contexts are set, which also prevent the freeze of the application:
QtCore.QCoreApplication.setAttribute(QtCore.Qt.AA_ShareOpenGLContexts)
# but don't want to use that since that needs to be set prior to an application initialization (which grill can't control as in USDView, Maya, Houdini...)
# https://stackoverflow.com/questions/56159475/qt-webengine-seems-to-be-initialized

# There's about ~0.4s overhead from creating a QApplication for the tests.

# 2024-11-09 - Python-3.13 & USD-24.11
# python -m unittest --durations 0 test_views
# Slowest test durations
# ----------------------------------------------------------------------
# 0.354s     test_spreadsheet_editor (tests.test_views.TestViews.test_spreadsheet_editor)
# 0.288s     test_connection_view (tests.test_views.TestViews.test_connection_view)
# 0.237s     test_taxonomy_editor (tests.test_views.TestViews.test_taxonomy_editor)
# 0.236s     test_content_browser (tests.test_views.TestViews.test_content_browser)
# 0.217s     test_scenegraph_composition (tests.test_views.TestViews.test_scenegraph_composition)
# 0.180s     test_dot_call (tests.test_views.TestViews.test_dot_call)
# 0.051s     test_prim_filter_data (tests.test_views.TestViews.test_prim_filter_data)
# 0.050s     test_create_assets (tests.test_views.TestViews.test_create_assets)
# 0.038s     test_stats (tests.test_views.TestViews.test_stats)
# 0.037s     test_graph_views (tests.test_views.TestViews.test_graph_views)
# 0.032s     test_prim_composition (tests.test_views.TestViews.test_prim_composition)
# 0.029s     test_display_color_editor (tests.test_views.TestViews.test_display_color_editor)
# 0.004s     test_pan (tests.test_views.TestViews.test_pan)
# 0.001s     test_horizontal_scroll (tests.test_views.TestViews.test_horizontal_scroll)
#
# (durations < 0.001s were hidden; use -v to show these durations)
# ----------------------------------------------------------------------
# Ran 18 tests in 2.141s


class TestPrivate(unittest.TestCase):
    def test_common_paths(self):
        input_paths = [
            Sdf.Path("/world/hi"),
            Sdf.Path.absoluteRootPath,
            Sdf.Path("/hola/hello/new1"),
            Sdf.Path("/world/child/nested"),
            Sdf.Path("/invalid/1"),
            Sdf.Path("/hola/hello/new2"),
            Sdf.Path("/hola/hello/new2/twochild"),
            Sdf.Path("/hola/hello/new2/twochild/more"),
            Sdf.Path("/hola/hello/new2/a"),
            Sdf.Path("/hola/hello/new2/zzzzzzzzzzzzzzzzzzz"),
            Sdf.Path("/hola/hello/new3"),
            Sdf.Path("/hola/hello/n9/nested/one"),
            Sdf.Path("/hola/hello/new01/nested/deep"),
            Sdf.Path("/hola/hello/n9/nested/two"),
            Sdf.Path("/hola/bye/child"),
            Sdf.Path("/deep/nested/unique/path"),
            Sdf.Path("/alone"),
        ]
        actual = usd.common_paths(input_paths)
        expected = [
            Sdf.Path('/alone'),
            Sdf.Path('/deep/nested/unique/path'),
            Sdf.Path('/hola/bye/child'),
            Sdf.Path('/hola/hello/n9/nested/one'),
            Sdf.Path('/hola/hello/n9/nested/two'),
            Sdf.Path('/hola/hello/new01/nested/deep'),
            Sdf.Path('/hola/hello/new1'),
            Sdf.Path('/hola/hello/new2'),
            Sdf.Path('/hola/hello/new3'),
            Sdf.Path('/world/child/nested'),
            Sdf.Path('/world/hi'),
        ]
        self.assertEqual(actual, expected)


_test_bed = Path(__file__).parent / "mini_test_bed" / "main-world-test.1.usda"


class TestViews(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls._app = QtWidgets.QApplication.instance() or QtWidgets.QApplication([])

    @classmethod
    def tearDownClass(cls):
        cls._app.quit()

    def test_connection_view(self):
        # https://openusd.org/release/tut_simple_shading.html
        stage = Usd.Stage.CreateInMemory()
        material = UsdShade.Material.Define(stage, '/TexModel/boardMat')
        pbrShader = UsdShade.Shader.Define(stage, '/TexModel/boardMat/PBRShader')
        roughness_name = "roughness"
        pbrShader.CreateInput(roughness_name, Sdf.ValueTypeNames.Float).Set(0.4)
        surface_name = "surface"
        material.CreateSurfaceOutput().ConnectToSource(pbrShader.ConnectableAPI(), surface_name)
        # Ensure cycles don't cause recursion
        cycle_input = pbrShader.CreateInput("cycle_in", Sdf.ValueTypeNames.Float)
        cycle_output = pbrShader.CreateOutput("cycle_out", Sdf.ValueTypeNames.Float)
        cycle_input.ConnectToSource(cycle_output)
        viewer = description._ConnectableAPIViewer()
        # GraphView capabilities are tested elsewhere, so mock 'view' here.
        viewer._graph_view.view = lambda indices: None
        viewer.setPrim(material)
        graph = viewer._graph_view._graph
        self.assertEqual(graph.nodes[str(material.GetPrim().GetPath())]['ports'], ['', surface_name])
        self.assertEqual(graph.nodes[str(pbrShader.GetPrim().GetPath())]['ports'], ['', cycle_input.GetBaseName(), roughness_name, cycle_output.GetBaseName(), surface_name])
        viewer.setPrim(None)

    def test_scenegraph_composition(self):
        """Confirm that bidirectionality between layer stacks completes.

        Bidirectionality in the composition graph is achieved by:
            - parent_stage -> child_stage via a reference, payload arcs
            - child_stage -> parent_stage via a inherits, specializes arcs
        """
        stage = Usd.Stage.Open(str(_test_bed))

        widget = description.LayerStackComposition()
        # GraphView capabilities are tested elsewhere, so mock 'view' here.
        widget._graph_view.view = lambda indices: None
        widget.setStage(stage)

        widget._layers.table.selectAll()
        expectedAffectedPrims = 306
        actualListedPrims = widget._prims.model.rowCount()
        self.assertEqual(expectedAffectedPrims, actualListedPrims)

        widget._graph_precise_source_ports.setChecked(True)
        widget._update_graph_from_graph_info(widget._computed_graph_info)

        widget._has_specs.setChecked(True)
        widget._graph_edge_include[description.Pcp.ArcTypeReference].setChecked(False)
        widget.setPrimPaths({"/Catalogue/Model/Elements/Apartment"})
        widget.setStage(stage)

        widget._layers.table.selectAll()
        self.assertEqual(5, widget._layers.model.rowCount())
        self.assertEqual(1, widget._prims.model.rowCount())

        _core._which.cache_clear()
        with mock.patch("grill.views.description._which") as patch:  # simulate dot is not in the environment
            patch.return_value = None
            widget._graph_view.view([0,1])

        _core._which.cache_clear()
        with mock.patch("grill.views.description.nx.nx_agraph.write_dot") as patch:  # simulate pydot not installed
            patch.side_effect = ImportError
            widget._graph_view.view([0])

        widget.deleteLater()

    def test_prim_composition(self):
        stage = Usd.Stage.Open(str(_test_bed))
        prim = stage.GetPrimAtPath("/Catalogue/Model/Buildings/Multi_Story_Building/Windows/Apartment/Geom/Floor")
        widget = description.PrimComposition()
        widget._include_inert_nodes.setChecked(False)
        widget._compute_expanded_index.setChecked(False)
        widget._complete_target_layerstack.setChecked(True)

        # DotView capabilities are tested elsewhere, so mock 'setDotPath' here.
        widget._dot_view.setDotPath = lambda fp: None
        widget.setPrim(prim)

        # cheap. prim is affected by 2 layers
        # single child for this prim.
        self.assertTrue(widget.composition_tree._model.invisibleRootItem().hasChildren())
        widget._compute_expanded_index.setChecked(True)
        widget._complete_target_layerstack.setChecked(True)
        widget.setPrim(prim)
        self.assertTrue(widget.composition_tree._model.invisibleRootItem().hasChildren())

        with mock.patch("grill.views.description.QtWidgets.QApplication.keyboardModifiers") as patch:
            patch.return_value = QtCore.Qt.ShiftModifier
            root_idx = widget.composition_tree._model.index(0, 0)
            widget.composition_tree.expanded.emit(root_idx)
            widget.composition_tree.collapsed.emit(root_idx)

        widget.setPrim(None)
        self.assertFalse(widget.composition_tree._model.invisibleRootItem().hasChildren())

        widget.clear()

    def test_spreadsheet_editor(self):
        widget = sheets.SpreadsheetEditor()
        widget._model_hierarchy.setChecked(False)  # default is True
        stage = Usd.Stage.Open(str(_test_bed))
        stage.SetEditTarget(stage.GetSessionLayer())
        with Sdf.ChangeBlock():
            stage.OverridePrim("/child_orphaned")
            stage.GetPrimAtPath("/Catalogue/Model/Blocks").SetActive(False)
        widget._orphaned.setChecked(True)
        widget.setStage(stage)
        self.assertEqual(stage, widget.stage)
        widget.table.scrollContentsBy(10, 10)

        widget.table.selectAll()

        widget.table.clearSelection()
        widget._column_options[0]._line_filter.setText("hade")
        widget._column_options[0]._updateMask()
        widget.table.resizeColumnToContents(0)

        widget.table.selectAll()
        expected_rows = {0, 1, 2}  # 1 prim from filtered name: /Catalogue/Shade /Catalogue/Shade/Color /Catalogue/Shade/Color/ModelDefault
        visible_rows = ({i.row() for i in widget.table.selectedIndexes()})
        self.assertEqual(visible_rows, expected_rows)

        widget._copySelection()
        clip = QtWidgets.QApplication.instance().clipboard().text()
        data = tuple(csv.reader(io.StringIO(clip), delimiter=csv.excel_tab.delimiter))
        expected_data = (
            ['/Catalogue/Shade', 'Shade', '', '', '', 'False', '', 'False'],
            ['/Catalogue/Shade/Color', 'Color', '', '', '', 'False', '', 'False'],
            ['/Catalogue/Shade/Color/ModelDefault', 'ModelDefault', 'ModelDefault', '', '', 'False', '', 'False'],
        )
        self.assertEqual(data, expected_data)

        widget.table.clearSelection()

        widget._model_hierarchy.click()  # enables model hierarchy, which we don't have any
        widget.table.selectAll()
        expected_rows = {0, 1}
        visible_rows = ({i.row() for i in widget.table.selectedIndexes()})
        self.assertEqual(expected_rows, visible_rows)

        widget.table.clearSelection()

        widget._lock_all.click()
        widget._conformLockSwitch()
        widget._vis_all.click()
        widget._conformVisibilitySwitch()

        widget._column_options[0]._line_filter.setText("")
        widget._model_hierarchy.click()  # disables model hierarchy, which we don't have any
        widget.table.selectAll()
        with mock.patch(f"{QtWidgets.__name__}.QMessageBox.warning", new=lambda *args: print(args)):
            widget._pasteClipboard()

        widget.model._prune_children = {Sdf.Path("/pruned")}

        widget._column_options[0]._line_filter.setText("")
        widget.table.clearSelection()
        widget._active.setChecked(False)
        widget._classes.setChecked(True)
        widget._filters_logical_op.setCurrentIndex(1)
        widget.stage = stage
        widget.table.selectAll()
        expected_colors = {str(each.value): each for each in sheets._PrimTextColor}  # colors are not hashable
        expected_fonts = {each.weight() for each in (  # font not hashable in PySide2
            sheets._prim_font(),
            sheets._prim_font(abstract=True),
            sheets._prim_font(abstract=True, orphaned=True),
            sheets._prim_font(orphaned=True),
        )}
        self.assertEqual(len(expected_fonts), 3)  # three weights: Light, ExtraLight, Normal
        collected_fonts = set()
        for each in widget.table.selectionModel().selectedIndexes():
            color_key = str(each.data(role=QtCore.Qt.ForegroundRole))
            font = each.data(role=QtCore.Qt.FontRole)
            font_key = font.weight()
            expected_colors.pop(color_key, None)
            collected_fonts.add(font_key)

        self.assertEqual(expected_fonts, collected_fonts)

    def test_prim_filter_data(self):
        stage = Usd.Stage.Open(str(_test_bed))
        stage.SetEditTarget(stage.GetSessionLayer())
        instance = stage.GetPrimAtPath('/Catalogue/Model/Blocks/Block_With_Inherited_Windows/Building1/Windows/Apartment')
        instance.SetActive(False)
        instance.Unload()
        over = stage.OverridePrim("/Orphaned")
        widget = sheets.SpreadsheetEditor()
        for stage_value in (stage, None):
            widget.setStage(stage_value)
            for each in range(2):
                widget._filters_logical_op.setCurrentIndex(each)
                widget._model_hierarchy.click()  # default is True
                widget._orphaned.click()
                widget._classes.click()
                widget._defined.click()
                widget._active.click()
                widget._inactive.click()
        widget.model._root_paths = {over.GetPath(), over.GetPath()}
        widget.model._prune_children = {over.GetPath()}
        widget.setStage(stage)

    def test_dot_call(self):
        """Test execution of function by mocking dot with python call"""
        with mock.patch("grill.views.description._which") as patch:
            patch.return_value = 'python'
            error, targetpath = _graph._dot_2_svg('nonexisting_path')
            # an error would be reported back
            self.assertIsNotNone(error)

    def test_content_browser(self):
        stage = Usd.Stage.Open(str(_test_bed))

        path_with_variant = Sdf.Path("/Origin{color=blue}Geom/Floor.primvars:displayColor")
        spawned_path = Sdf.Path("/Catalogue/Model/Buildings/Multi_Story_Building/Windows/Apartment")
        apartments_layer = Sdf.Layer.FindOrOpen(str(_test_bed.parent / "Model-Elements-Apartment.1.usda"))
        layers = stage.GetLayerStack() + [apartments_layer]
        args = layers, None, stage.GetPathResolverContext(), (Sdf.Path("/"), spawned_path, path_with_variant)
        anchor = layers[0]

        _core_run = _core._run

        def _fake_run(run_args: list):
            return "", Sdf.Layer.FindOrOpen(run_args[-1]).ExportToString()

        # sdffilter still not coming via pypi, so patch for now
        with mock.patch("grill.views.description._core._run", new=_fake_run):
            dialog = description._start_content_browser(*args)
            browser: description._PseudoUSDBrowser = dialog.findChild(description._PseudoUSDBrowser)
            assert browser._browsers_by_layer.values()
            first_browser_widget, = browser._browsers_by_layer.values()
            first_browser_widget._format_options.setCurrentIndex(0)  # pseudoLayer (through sdffilter)
            first_browser_widget._format_options.setCurrentIndex(1)  # outline (through sdffilter)
            first_browser_widget._format_options.setCurrentIndex(2)  # usdtree (through usdtree)
            first_browser_widget._format_options.setCurrentIndex(0)

            browser_tab: description._PseudoUSDTabBrowser = first_browser_widget.findChild(description._PseudoUSDTabBrowser)
            browser._on_identifier_requested(anchor, layers[1].identifier)
            with mock.patch(f"{QtWidgets.__name__}.QMessageBox.warning", new=lambda *args: print(args)):
                browser._on_identifier_requested(anchor, "/missing/file.usd")
                _, empty_png = tempfile.mkstemp(suffix=".png")
                browser._on_identifier_requested(anchor, empty_png)
                _, empty_usd = tempfile.mkstemp(suffix=".usda")
                browser._on_identifier_requested(anchor, empty_usd)

            menu = browser._menu_for_tab(0)
            self.assertTrue(bool(menu.actions()))

            position = QtCore.QPoint(10, 10)
            pixelDelta = QtCore.QPoint(0, 0)
            angleDelta_zoomIn = QtCore.QPoint(0, 120)
            buttons = QtCore.Qt.NoButton
            modifiers = QtCore.Qt.ControlModifier
            phase = QtCore.Qt.NoScrollPhase
            inverted = False

            # ZOOM IN
            event = QtGui.QWheelEvent(position, position, pixelDelta, angleDelta_zoomIn, buttons, modifiers, phase, inverted)
            browser_tab.wheelEvent(event)

            # Assert that the scale has changed according to the zoom logic
            angleDelta_zoomOut = QtCore.QPoint(-120, 0)

            # ZOOM OUT
            event = QtGui.QWheelEvent(position, position, pixelDelta, angleDelta_zoomOut, buttons, modifiers, phase, inverted)
            browser_tab.wheelEvent(event)

            browser._close_many(range(len(browser._tab_layer_by_idx)))
            for child in dialog.findChildren(description._PseudoUSDBrowser):
                child._resolved_layers.clear()

            # create a temporary file loadable by our image tab
            image = QtGui.QImage(QtCore.QSize(1, 1), QtGui.QImage.Format_RGB888)
            image.fill(QtGui.QColor(255, 0, 0))
            targetpath = str(_test_bed.with_suffix(".jpg"))
            image.save(targetpath, "JPG")
            browser._on_identifier_requested(anchor, targetpath)

            invalid_crate_layer = Sdf.Layer.CreateAnonymous()
            invalid_crate_layer.ImportFromString(
                # Not valid in USD-24.05: https://github.com/PixarAnimationStudios/OpenUSD/blob/59992d2178afcebd89273759f2bddfe730e59aa8/pxr/usd/sdf/testenv/testSdfParsing.testenv/baseline/127_varyingRelationship.sdf#L9
                """#sdf 1.4.32
                def GprimSphere "Sphere"
                {
                    delete varying rel constraintTarget = </Pivot3>
                    add varying rel constraintTarget = [
                        </Pivot3>,
                        </Pivot2>,
                    ]
                    reorder varying rel constraintTarget = [
                        </Pivot2>,
                        </Pivot>,
                    ]
                    varying rel constraintTarget.default = </Pivot>    
                }
                """
            )
            description._start_content_browser([invalid_crate_layer], None, stage.GetPathResolverContext(), ())

        with mock.patch("grill.views.description._which") as patch:
            patch.return_value = None
            with self.assertRaisesRegex(ValueError, "Expected arguments to contain an executable value on the first index"):
                description._start_content_browser(*args)

        error, result = _core._run(["python", 42])
        self.assertTrue(error.startswith('expected str'))
        self.assertEqual(result, "")

    def test_display_color_editor(self):
        stage = Usd.Stage.Open(str(_test_bed))
        stage.SetEditTarget(stage.GetSessionLayer())
        sphere = UsdGeom.Sphere(stage.GetPrimAtPath('/Catalogue/Model/Elements/Apartment/Geom/Volume'))
        color_var = sphere.GetDisplayColorPrimvar()
        editor = _attributes._DisplayColorEditor(color_var)

        with mock.patch("grill.views._attributes.QtWidgets.QColorDialog.getColor", new=lambda *_, **__: QtGui.QColor(255, 255, 0)):
            editor._color_launchers["Color"][0].click()

        color_var.SetInterpolation(UsdGeom.Tokens.vertex)
        editor = _attributes._DisplayColorEditor(color_var)
        editor._update_value()
        editor._random.click()

        xform = stage.DefinePrim("/x")
        primvar = UsdGeom.Gprim(xform.GetPrim()).CreateDisplayColorPrimvar()
        editor = _attributes._DisplayColorEditor(primvar)
        with self.assertRaises(TypeError):  # atm some gprim types are not supported
            editor._update_value()

        editor = _attributes._DisplayColorEditor(UsdGeom.Primvar())
        self.assertEqual(len(editor._value), 1)

    def test_stats(self):
        empty = stats.StageStats()
        self.assertEqual(empty._usd_tree.topLevelItemCount(), 0)

        stage = Usd.Stage.Open(str(_test_bed))
        widget = stats.StageStats(stage=stage)
        self.assertGreater(widget._usd_tree.topLevelItemCount(), 1)
        current = _qt.QtCharts
        del _qt.QtCharts
        stats.StageStats(stage=stage)
        _qt.QtCharts = current

    def test_graph_views(self):
        viewer = _graph.GraphView()

        with (
            mock.patch(f"grill.views._graph.drawing.nx_pydot.graphviz_layout", new=lambda graph, **__: dict.fromkeys(graph.nodes, (0,0))),
        ):
            for invalid_node_data, error_message in (
                    (dict(shape='record'), "'label' must be supplied"),
                    (dict(shape='record', label='no record'), "a record 'label' in the form of"),
                    (dict(shape='record', label='{1}'), "a record 'label' in the form of"),
                    (dict(shape='record', label='{<0>1}', ports=('first', 'second')), "record 'shape' and 'ports' are mutually exclusive"),
                    (dict(shape='none'), "A label must be provided"),
            ):
                invalid_graph = _graph.nx.MultiDiGraph()
                invalid_graph.add_nodes_from([(1, invalid_node_data)])
                with self.assertRaisesRegex(ValueError, error_message):
                    viewer.graph = invalid_graph

        viewer = _graph.GraphView()
        viewer.view(tuple())

        nodes_info = {
            1: dict(
                label="{<one>x:y:z|<two>z}",
                style="rounded",  # these can be set at the graph level
                shape="record",
            ),
            2: dict(
                label="{<three>a|<four>b}",
                style='rounded',
                shape="record",
            ),
            3: dict(
                label="{<five>c|<six>d}",
                style='rounded',
                shape="record",
            ),
            "parent": dict(
                shape="box", fillcolor="#afd7ff", color="#1E90FF", style="filled,rounded"
            ),
            "child1": dict(
                shape="box", fillcolor="#afd7ff", color="#1E90FF", style="filled,rounded"
            ),
            "child2": dict(
                shape="box", fillcolor="#afd7ff", color="#1E90FF", style="invis"
            ),
        }
        edges_info = (
            (1, 1, dict(color='sienna:crimson:orange')),
            (1, 2, dict(color='crimson')),
            (2, 1, dict(color='seagreen')),
            (3, 2, dict(color='steelblue', tailport='five')),
            (3, 1, dict(color='hotpink', tailport='five')),
            ("parent", "child1"),
            ("parent", "child2", dict(label='invis')),
        )

        graph = _graph.nx.MultiDiGraph()
        graph.add_nodes_from(nodes_info.items())
        graph.add_edges_from(edges_info)
        graph.graph['graph'] = {'rankdir': 'LR'}
        graph.graph['edge'] = {"color": 'crimson'}
        outline_color = "#4682B4"  # 'steelblue'
        background_color = "#F0FFFF"  # 'azure'
        table_row = '<tr><td port="{port}" border="0" bgcolor="{color}" style="ROUNDED">{text}</td></tr>'

        connection_nodes = dict(
            ancestor=dict(
                ports=('', 'cycle_in', 'roughness', 'cycle_out', 'surface'),
                shape='none',
                connections=dict(
                    surface=[('successor', 'surface')],
                    cycle_out=[('ancestor', 'cycle_in')],
                ),
            ),
            successor=dict(
                ports=('', 'surface'),
                shape='none',
                connections=dict(),
            )
        )
        connection_edges = []

        def _add_edges(src_node, src_name, tgt_node, tgt_name):
            tooltip = f"{src_node}.{src_name} -> {tgt_node}.{tgt_name}"
            connection_edges.append((src_node, tgt_node, {"tailport": src_name, "headport": tgt_name, "tooltip": tooltip}))

        for node, data in connection_nodes.items():
            label = f'<<table border="1" cellspacing="2" style="ROUNDED" bgcolor="{background_color}" color="{outline_color}">'
            label += table_row.format(port="", color="white",
                                      text=f'<font color="{outline_color}"><b>{node}</b></font>')
            for port in data['ports']:
                if not port:
                    continue
                sources = data['connections'].get(port, [])  # (valid, invalid): we care only about valid sources (index 0)
                color = r"#F08080" if sources else background_color
                label += table_row.format(port=port, color=color, text=f'<font color="#242828">{port}</font>')
                for source_node, source_port in sources:
                    # node_id='ancestor', port_name='cycle_out', ancestor, source.sourceName='cycle_in'
                    # tooltip='/TexModel/boardMat/PBRShader.cycle_in -> /TexModel/boardMat/PBRShader.cycle_out'
                    _add_edges(node, port, source_node, source_port)

            label += '</table>>'
            data['label'] = label
            data.pop('connections', None)

        graph.add_nodes_from(connection_nodes.items())
        graph.add_edges_from(connection_edges)

        widget = QtWidgets.QFrame()
        splitter = QtWidgets.QSplitter(QtCore.Qt.Horizontal)

        def _use_test_dot(subgraph, fp):
            source_path = Path(__file__).parent / "test_data/_mini_graph.dot"
            fp.write(source_path.read_text(encoding="utf-8"))

        def _use_test_svg(self, filepath):
            return self._on_dot_result(str(Path(__file__).parent / "test_data/_mini_graph.svg"))

        def _test_positions(graph, prog):
            return {
                1: (218.75, 90.1),
                2: (322.75, 90.1),
                3: (76.125, 61.1),
                'parent': (76.125, 190.1),
                'child1': (218.75, 217.1),
                'child2': (218.75, 163.1),
                'ancestor': (76.125, 316.1),
                'successor': (218.75, 282.1),
            }

        with (
            mock.patch(f"grill.views._graph.nx.nx_pydot.write_dot", new=_use_test_dot),
            mock.patch(f"grill.views._graph.nx.nx_agraph.write_dot", new=_use_test_dot),
            mock.patch(f"grill.views._graph._DotViewer.setDotPath", new=_use_test_svg),
            mock.patch(f"grill.views._graph.drawing.nx_pydot.graphviz_layout", new=_test_positions),
        ):
            for cls in _graph.GraphView, _graph._GraphSVGViewer:
                for pixmap_enabled in ((True, False) if cls == _graph._GraphSVGViewer else (False,)):
                    _graph._USE_SVG_VIEWPORT = pixmap_enabled

                    child = cls(parent=widget)

                    if isinstance(child, _graph.GraphView):
                        with self.assertRaisesRegex(LookupError, "Could not find sender"):
                            invalid_uril = QtCore.QUrl(f"{child.url_id_prefix}not_a_digit")
                            child._graph_url_changed(invalid_uril)
                    else:
                        with self.assertRaisesRegex(RuntimeError, "'graph' attribute not set yet"):
                            invalid_uril = QtCore.QUrl(f"{child.url_id_prefix}not_a_digit")
                            child._graph_url_changed(invalid_uril)
                        child._on_dot_error("nothing set yet")

                    if cls == _graph._GraphSVGViewer and not pixmap_enabled:
                        # QWebEngineView in use, no need to test its 'load' method
                        child._graph_view.load = lambda fp: None
                    child._graph = graph
                    child.view(graph.nodes)
                    child.setMinimumWidth(150)
                    splitter.addWidget(child)

                    if isinstance(child, _graph.GraphView):
                        for item in child.scene().items():
                            item.boundingRect()  # trigger bounding rect logic
                            if isinstance(item, _graph._Edge):
                                cycle_collected = True
                            if isinstance(item, _graph._Node):
                                nodes_hovered_checked = True

                                # Test hover with no modifiers
                                event = QtWidgets.QGraphicsSceneHoverEvent(QtCore.QEvent.GraphicsSceneHoverMove)
                                center = item.sceneBoundingRect().center()
                                event.setScenePos(center)
                                item.hoverEnterEvent(event)
                                self.assertEqual(item.cursor().shape(), QtGui.Qt.ArrowCursor)
                                self.assertEqual(item.textInteractionFlags(), item._default_text_interaction)
                                item.hoverLeaveEvent(event)

                                # Test hover with Ctrl modifier
                                event = QtWidgets.QGraphicsSceneHoverEvent(QtCore.QEvent.GraphicsSceneHoverMove)
                                event.setScenePos(center)
                                event.setModifiers(QtCore.Qt.ControlModifier)
                                item.hoverEnterEvent(event)
                                self.assertEqual(item.cursor().shape(), QtGui.Qt.PointingHandCursor)
                                item.hoverLeaveEvent(event)

                                # Test hover with Alt modifier
                                event = QtWidgets.QGraphicsSceneHoverEvent(QtCore.QEvent.GraphicsSceneHoverMove)
                                event.setScenePos(item.sceneBoundingRect().center())
                                event.setModifiers(QtCore.Qt.AltModifier)
                                item.hoverEnterEvent(event)
                                self.assertEqual(item.cursor().shape(), QtGui.Qt.ClosedHandCursor)
                                self.assertEqual(item.textInteractionFlags(), QtCore.Qt.NoTextInteraction)
                                item.hoverLeaveEvent(event)

                                item.itemChange(QtWidgets.QGraphicsItem.ItemPositionHasChanged, (1,1))

                        self.assertTrue(cycle_collected)
                        self.assertTrue(nodes_hovered_checked)

                    child.filter_edges = lambda src, tgt, port: src not in graph.nodes
                    child.view(graph.nodes)

    def test_zoom(self):
        """Zoom is triggered by ctrl + mouse wheel"""
        view = _graph._GraphicsViewport()

        initial_scale = view.transform().m11()

        position = QtCore.QPoint(10, 10)
        pixelDelta = QtCore.QPoint(0, 0)
        angleDelta_zoomIn = QtCore.QPoint(0, 120)
        buttons = QtCore.Qt.NoButton
        modifiers = QtCore.Qt.ControlModifier
        phase = QtCore.Qt.NoScrollPhase
        inverted = False

        # ZOOM IN
        event = QtGui.QWheelEvent(position, position, pixelDelta, angleDelta_zoomIn, buttons, modifiers, phase, inverted)
        view.wheelEvent(event)

        zoomed_in_scale = view.transform().m11()

        # Assert that the scale has changed according to the zoom logic
        self.assertGreater(zoomed_in_scale, initial_scale)

        angleDelta_zoomOut = QtCore.QPoint(-120, 0)

        # ZOOM OUT
        event = QtGui.QWheelEvent(position, position, pixelDelta, angleDelta_zoomOut, buttons, modifiers, phase, inverted)
        view.wheelEvent(event)
        self.assertGreater(zoomed_in_scale, view.transform().m11())

    def test_horizontal_scroll(self):
        """Horizontal scrolling with alt + mouse wheel"""
        view = _graph._GraphicsViewport()
        scroll_bar = view.horizontalScrollBar()
        initial_value = scroll_bar.value()
        scroll_bar.setMaximum(200)
        position = QtCore.QPoint(10, 10)
        pixelDelta = QtCore.QPoint(0, 0)
        angleDelta= QtCore.QPoint(-120, 0)
        buttons = QtCore.Qt.NoButton
        modifiers = QtCore.Qt.AltModifier
        phase = QtCore.Qt.NoScrollPhase
        inverted = False
        event = QtGui.QWheelEvent(position, position, pixelDelta, angleDelta, buttons, modifiers, phase, inverted)
        view.wheelEvent(event)
        final_value = scroll_bar.value()
        # Assert that the horizontal scroll has changed according to your pan logic
        self.assertGreater(final_value, initial_value)

    def test_vertical_scroll(self):
        """Vertical scroll with only mouse wheel"""
        view = _graph._GraphicsViewport()
        scroll_bar = view.verticalScrollBar()
        initial_value = scroll_bar.value()
        scroll_bar.setMaximum(200)
        position = QtCore.QPoint(10, 10)
        pixelDelta = QtCore.QPoint(0, 0)
        angleDelta = QtCore.QPoint(0, -120)
        buttons = QtCore.Qt.NoButton
        modifiers = QtCore.Qt.NoModifier
        phase = QtCore.Qt.NoScrollPhase
        inverted = False
        event = QtGui.QWheelEvent(position, position, pixelDelta, angleDelta, buttons, modifiers, phase, inverted)
        view.wheelEvent(event)
        final_value = scroll_bar.value()
        # Assert that the horizontal scroll has changed according to your pan logic
        self.assertGreater(final_value, initial_value)

    def test_pan(self):
        """Horizontal and vertical pan with mouse middle button"""
        view = _graph._GraphicsViewport()
        vertical_scroll_bar = view.verticalScrollBar()
        vertical_scroll_bar.setMaximum(200)
        horizontal_scroll_bar = view.horizontalScrollBar()
        horizontal_scroll_bar.setMaximum(200)
        start_position = QtCore.QPoint(50, 50)
        end_position = QtCore.QPoint(-5, -5)

        # 1. Mouse press
        middle_button_event = QtGui.QMouseEvent(
            QtCore.QEvent.MouseButtonPress,
            start_position,
            start_position,
            QtCore.Qt.MiddleButton,
            QtCore.Qt.MiddleButton,
            QtCore.Qt.NoModifier,
        )
        vertical_value = vertical_scroll_bar.value()
        horizontal_value = horizontal_scroll_bar.value()
        view.mousePressEvent(middle_button_event)
        self.assertEqual(self._app.overrideCursor().shape(), QtGui.Qt.ClosedHandCursor)
        self.assertTrue(view._dragging)

        # 2. Mouse move
        view._last_pan_pos = _graph._EVENT_POSITION_FUNC(middle_button_event) + QtCore.QPoint(10,10)
        move_event = QtGui.QMouseEvent(
            QtCore.QEvent.MouseButtonPress,
            end_position,
            end_position,
            QtCore.Qt.MiddleButton,
            QtCore.Qt.MiddleButton,
            QtCore.Qt.NoModifier,
        )
        view.mouseMoveEvent(move_event)
        last_vertical_scroll_bar = vertical_scroll_bar.value()
        last_horizontal_scroll_bar = horizontal_scroll_bar.value()
        self.assertGreater(last_vertical_scroll_bar, vertical_value)
        self.assertGreater(last_horizontal_scroll_bar, horizontal_value)

        # 3. Release
        view.mouseReleaseEvent(middle_button_event)
        view._last_pan_pos = _graph._EVENT_POSITION_FUNC(middle_button_event) + QtCore.QPoint(20, 20)
        view.mouseMoveEvent(move_event)
        # Confirm no further move is performed
        self.assertEqual(last_vertical_scroll_bar, vertical_scroll_bar.value())
        self.assertEqual(last_horizontal_scroll_bar, horizontal_scroll_bar.value())


@unittest.skipUnless(_COOK_AVAILABLE, "Unable to test without 'grill.cook' module")
class TestCreationViews(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls._app = QtWidgets.QApplication.instance() or QtWidgets.QApplication([])

    def setUp(self):
        self._tmpf = tempfile.mkdtemp()
        self._token = cook.Repository.set(cook.Path(self._tmpf) / "repo")

    def tearDown(self) -> None:
        cook.Repository.reset(self._token)
        shutil.rmtree(self._tmpf)

    def test_taxonomy_editor(self):
        class MiniAsset(names.UsdAsset):
            drop = ('code', 'media', 'area', 'stream', 'step', 'variant', 'part')
            DEFAULT_SUFFIX = "usda"

        cook.UsdAsset = MiniAsset
        stage = Usd.Stage.Open(str(_test_bed))
        existing = list(cook.itaxa(stage))
        widget = create.TaxonomyEditor()
        # GraphView capabilities are tested elsewhere, so mock 'view' here.
        widget._graph_view.view = lambda indices: None

        widget.setStage(stage)

        widget._amount.setValue(3)  # TODO: create 10 assets, clear tmp directory

        valid_data = (
            ['NewType1', existing[0].GetName(), 'Id1', ],
            ['NewType2', '', 'Id2', ],
        )
        data = valid_data + (
            ['',         existing[0].GetName(), 'Id3', ],
        )

        QtWidgets.QApplication.instance().clipboard().setText('')
        widget.sheet._pasteClipboard()

        stream = io.StringIO()
        csv.writer(stream, delimiter=csv.excel_tab.delimiter).writerows(data)
        QtWidgets.QApplication.instance().clipboard().setText(stream.getvalue())

        widget.sheet.table.selectAll()
        widget.sheet._pasteClipboard()
        widget._create()

        for name, __, __ in valid_data:
            created = stage.GetPrimAtPath(cook._TAXONOMY_ROOT_PATH).GetPrimAtPath(name)
            self.assertTrue(created.IsValid())

        sheet_model = widget.sheet.model
        index = sheet_model.index(0, 1)
        editor = widget.sheet._columns[1].editor(None, None, index)
        self.assertIsInstance(editor, QtWidgets.QDialog)
        widget.sheet._columns[1].setter(editor, sheet_model, index)
        editor._options.selectAll()
        menu = editor._create_context_menu()
        menu.actions()[0].trigger()

        # after creation, set stage again to test existing column
        widget._apply()
        widget._existing.table.selectAll()
        selected_items = widget._existing.table.selectedIndexes()
        self.assertEqual(len(selected_items), len(valid_data) + len(existing))
        stage.Reload()  # We modified the taxonomy of the testbed. Reload to discard changes.

    def test_create_assets(self):
        stage = cook.fetch_stage(cook.UsdAsset.get_anonymous())
        for each in range(1, 6):
            cook.define_taxon(stage, f"Option{each}")

        widget = create.CreateAssets()
        widget.setStage(stage)

        widget._amount.setValue(3)  # TODO: create 10 assets, clear tmp directory

        data = (
            ['Option1', 'asset01', 'Asset 01', 'Description 01'],
            ['Option2', 'asset02', 'Asset 02', 'Description 02'],
            ['Option2', '',        'Asset 03', 'Description 03'],
        )

        QtWidgets.QApplication.instance().clipboard().setText('')
        widget.sheet._pasteClipboard()

        stream = io.StringIO()
        csv.writer(stream, delimiter=csv.excel_tab.delimiter).writerows(data)
        QtWidgets.QApplication.instance().clipboard().setText(stream.getvalue())

        widget.sheet.table.selectAll()
        widget.sheet._pasteClipboard()
        widget._create()
        taxon_editor = widget.sheet._columns[0].editor(widget, None, None)
        self.assertIsInstance(taxon_editor, QtWidgets.QComboBox)
        widget._apply()



import uuid
import logging
import unittest
import tempfile

from pathlib import Path

from pxr import Usd, UsdGeom, Sdf, Ar, Tf

try:
    from grill import cook, names, tokens, usd
    _COOK_AVAILABLE = True
except ImportError as exc:
    _COOK_AVAILABLE = False
    print(f"'grill.cook' module failed to import. Unable to test: {exc.msg}")

logger = logging.getLogger(__name__)

# 2024-11-09 - Python-3.13 & USD-24.11
# python -m unittest --durations 0 test_cook
# Slowest test durations
# ----------------------------------------------------------------------
# 0.044s     test_inherited_and_specialized_contexts (test_cook.TestCook.test_inherited_and_specialized_contexts)
# 0.036s     test_asset_unit (test_cook.TestCook.test_asset_unit)
# 0.024s     test_spawn_many (test_cook.TestCook.test_spawn_many)
# 0.024s     test_create_many_in_memory (test_cook.TestCook.test_create_many_in_memory)
# 0.014s     test_taxonomy (test_cook.TestCook.test_taxonomy)
# 0.010s     test_fetch_stage (test_cook.TestCook.test_fetch_stage)
#
# ----------------------------------------------------------------------
# Ran 6 tests in 0.152s

@unittest.skipUnless(_COOK_AVAILABLE, "Unable to test without 'grill.cook' module")
class TestCook(unittest.TestCase):
    def setUp(self) -> None:
        tempdir = tempfile.mkdtemp()
        logger.debug(f"Repository root directory: {tempdir}")
        self.root_asset = names.UsdAsset.get_anonymous()
        self.token = cook.Repository.set(Path(tempdir) / "repo")

    def tearDown(self) -> None:
        cook.Repository.reset(self.token)

    def test_fetch_stage(self):
        root_asset = self.root_asset
        root_stage = cook.fetch_stage(root_asset)

        # fetching stage outside of AR _context should resolve to same stage
        self.assertEqual(cook.asset_identifier(root_stage.GetRootLayer().identifier), root_asset.name)

        repo_path = cook.Repository.get()
        resolver_ctx = Ar.DefaultResolverContext([str(repo_path)])

        # confirm that a stage from a layer created separately is fetched with the correct resolver context
        usd_opened = str(names.UsdAsset.get_anonymous(item='usd_opened'))
        Sdf.Layer.CreateNew(str(repo_path / usd_opened))

        with self.assertRaisesRegex(Tf.ErrorException, "Failed to open layer"):
            # no resolver context, so unable to open stage
            Usd.Stage.Open(usd_opened)

        with Ar.ResolverContextBinder(resolver_ctx):
            opened_stage = Usd.Stage.Open(usd_opened)

        # when fetching the same asset identifier, the root layer should be the same
        fetched_stage = cook.fetch_stage(usd_opened)
        self.assertIs(opened_stage.GetRootLayer(), fetched_stage.GetRootLayer())
        # but the stages should be different
        self.assertIsNot(opened_stage, fetched_stage)

        in_memory = Usd.Stage.CreateInMemory()
        from_memory = str(names.UsdAsset.get_anonymous(item='from_memory'))
        with self.assertRaises(ValueError):
            # a stage with an empty resolver that fetches a valid identifier should fail.
            cook.fetch_stage(from_memory, context=in_memory.GetPathResolverContext())

    def test_taxonomy(self):
        stage = Usd.Stage.CreateInMemory()

        # An anonymous stage (non grill anonymous) should fail to define taxon.
        with self.assertRaisesRegex(ValueError, "Could not find a valid pipeline layer"):
            cook.define_taxon(stage, "ShouldFail")

        # Same stage containing a grill anon layer on its stack should succeed.
        anon_pipeline = Sdf.Layer.CreateNew(str(cook.Repository.get() / names.UsdAsset.get_anonymous().name))
        stage.GetRootLayer().subLayerPaths.append(anon_pipeline.realPath)

        # Now, test stages fetched from the start via "common" pipeline calls.
        with self.assertRaisesRegex(ValueError, "reserved name"):
            cook.define_taxon(stage, cook._TAXONOMY_NAME)

        with self.assertRaisesRegex(ValueError, "must be a valid identifier for a prim"):
            cook.define_taxon(stage, "/InvalidName")

        with self.assertRaisesRegex(ValueError, "reserved id fields"):
            cook.define_taxon(stage, "taxonomy_not_allowed", id_fields={cook._TAXONOMY_UNIQUE_ID: "by_id_value"})

        with self.assertRaisesRegex(ValueError, "reserved id fields"):
            cook.define_taxon(stage, "taxonomy_not_allowed", id_fields={cook._TAXONOMY_UNIQUE_ID.name: "by_id_name"})

        with self.assertRaisesRegex(ValueError, "invalid id_field keys"):
            cook.define_taxon(stage, "nonexistingfield", id_fields={str(uuid.uuid4()): "by_id_name"})

        missing_or_empty_fields_msg = f"Missing or empty '{cook._FIELDS_KEY}'"
        not_taxon = stage.DefinePrim("/not/a/taxon")
        with self.assertRaisesRegex(ValueError, missing_or_empty_fields_msg):
            cook.create_unit(not_taxon, "NoTaxon")

        not_taxon.SetAssetInfoByKey(cook._ASSETINFO_KEY, {})
        with self.assertRaisesRegex(ValueError, missing_or_empty_fields_msg):
            cook.create_unit(not_taxon, "EmptyAssetInfo")

        not_taxon.SetAssetInfoByKey(cook._ASSETINFO_KEY, {'invalid': 42})
        with self.assertRaisesRegex(ValueError, missing_or_empty_fields_msg):
            cook.create_unit(not_taxon, "InvalidAssetInfo")

        not_taxon.SetAssetInfoByKey(cook._ASSETINFO_KEY, {cook._FIELDS_KEY: 42})
        with self.assertRaisesRegex(TypeError, f"Expected mapping on key '{cook._FIELDS_KEY}'"):
            cook.create_unit(not_taxon, "InvalidAssetInfo")

        not_taxon.SetAssetInfoByKey(cook._ASSETINFO_KEY, {cook._FIELDS_KEY: {}})
        with self.assertRaisesRegex(ValueError, missing_or_empty_fields_msg):
            cook.create_unit(not_taxon, "EmptyFields")

        first = cook.define_taxon(stage, "first")
        # idempotent call should keep previously created prim
        self.assertEqual(first, cook.define_taxon(stage, "first"))

        second = cook.define_taxon(stage, "second", references=(first,))
        self.assertTrue(second.IsValid())

        third = cook.define_taxon(stage, "third", references=(first,))

        found_taxa = set(cook.itaxa(stage))
        self.assertSetEqual(set(found_taxa), {first, second, third})

        with self.assertRaisesRegex(ValueError, "is not a taxon."):
            cook.taxonomy_graph([first.GetParent()], "")

        graph_from_stage = cook.taxonomy_graph(found_taxa, "")
        first_successors = set(graph_from_stage.successors(first.GetName()))
        self.assertEqual(first_successors, {second.GetName(), third.GetName()})
        self.assertEqual(set(cook.taxonomy_graph(stage, "").nodes), set(graph_from_stage.nodes))

    def test_filter_taxa(self):
        stage = cook.fetch_stage(self.root_asset)
        root = cook.define_taxon(stage, "Root")
        parent, child = cook.create_many(root, ['A', 'B'])
        another = cook.define_taxon(stage, "Another")
        cook.create_many(another, ['C', 'D'])
        inherited = cook.define_taxon(stage, "Inherited", references=[root])
        grandchild = cook.create_unit(inherited, 'E')
        self.assertSetEqual({parent, child, grandchild}, set(cook.filter_taxa(stage.Traverse(), root)))

    def test_asset_unit(self):
        stage = cook.fetch_stage(self.root_asset)
        taxon_name = "taxon"
        unit_name = "unit"
        unit = cook.create_unit(cook.define_taxon(stage, "taxon"), "unit")
        unit_asset = cook.unit_asset(unit)
        unit_id = names.UsdAsset(cook.asset_identifier(unit_asset.identifier))
        self.assertEqual(unit_name, getattr(unit_id, cook._UNIT_UNIQUE_ID.name))
        self.assertEqual(taxon_name, getattr(unit_id, cook._TAXONOMY_UNIQUE_ID.name))

        # When asset identifier is empty or non existing, the fallback inspection of the prim itself should get the same result
        Usd.ModelAPI(unit).SetAssetIdentifier("")
        self.assertEqual(cook.unit_asset(unit).identifier, unit_asset.identifier)

    def test_create_many_in_memory(self):
        stage = Usd.Stage.CreateInMemory()
        # Root_asset is an empty anonymous asset with a pipeline compliant identifier. Create and sublayer it
        anon_pipeline = Sdf.Layer.CreateNew(str(cook.Repository.get() / self.root_asset.name))
        stage.GetRootLayer().subLayerPaths.append(anon_pipeline.identifier)
        cook.create_many(cook.define_taxon(stage, "Anon"), ("first", "second"))

    def test_spawn_many(self):
        stage = cook.fetch_stage(self.root_asset)

        parent = stage.DefinePrim("/a")
        with self.assertRaisesRegex(ValueError, "Can not spawn .* to itself."):
            cook.spawn_many(parent, parent, ["impossible"])
        child = stage.DefinePrim("/b")  # child needs to be a grill unit
        with self.assertRaisesRegex(ValueError, "Could not extract identifier from"):
            cook.spawn_many(parent, child, ["b"])
        invalid_path = "/invalid/path"
        with self.assertRaisesRegex(ValueError, "needs to be a child path of parent path"):
            cook.spawn_unit(parent, child, invalid_path)

        id_fields = {tokens.ids.CGAsset.kingdom.name: "K"}
        taxon = cook.define_taxon(stage, "Another", id_fields=id_fields)
        parent, child = cook.create_many(taxon, ['A', 'B'])
        cook.spawn_many(parent, child, ["b", "nested/c"], labels=["1", "2", "3"])
        self.assertEqual(len(parent.GetChildren()), 2)

        geom = cook.fetch_stage(cook.UsdAsset.get_anonymous(part="Geom"))
        geom.SetDefaultPrim(geom.DefinePrim(cook._UNIT_ORIGIN_PATH))
        payload = Sdf.Payload(cook.asset_identifier(geom.GetRootLayer().identifier))

        with cook.unit_context(child):
            geom_root = stage.DefinePrim(child.GetPath().AppendChild("Geom"))
            geom_root.GetPayloads().AddPayload(payload)
            with usd.edit_context(payload, geom_root):
                with self.assertRaisesRegex(RuntimeError, 'No spec path.*Could not set kind to "assembly"'):
                    cook.spawn_many(child, parent, paths=[geom_root.GetPath().AppendChild('this_will_fail')])

    def test_inherited_and_specialized_contexts(self):
        stage = cook.fetch_stage(self.root_asset)
        id_fields = {tokens.ids.CGAsset.kingdom.name: "K"}
        taxon = cook.define_taxon(stage, "Another", id_fields=id_fields)
        parent_unit, to_be_specialized, to_be_inherited, not_under_context = cook.create_many(
            taxon, ['parent_unit', 'to_be_specialized', 'to_be_inherited', 'not_under_context']
        )

        not_a_unit = stage.DefinePrim("/vanilla_prim")

        with self.assertRaisesRegex(ValueError, "is not a valid unit"):
            cook.specialized_context(not_a_unit)
        with self.assertRaisesRegex(ValueError, "needs to be a valid unit"):
            # current parent prim is not a valid unit in the catalogue (it's just a group)
            cook.specialized_context(to_be_specialized, to_be_specialized.GetParent())

        with self.assertRaisesRegex(ValueError, "is not a descendant"):
            cook.specialized_context(parent_unit, to_be_specialized)
        # return
        spawned_invalid = cook.spawn_unit(parent_unit, not_under_context)

        with self.assertRaisesRegex(ValueError, "Is there a composition arc bringing"):
            # TODO: find a more meaningful message (higher level) than the edit target context one.
            cook.specialized_context(spawned_invalid, parent_unit)

        with cook.unit_context(parent_unit):
            specialized_spawned = cook.spawn_unit(parent_unit, to_be_specialized)
            inherited_spawned = cook.spawn_unit(parent_unit, to_be_inherited)

        with cook.inherited_context(parent_unit):
            UsdGeom.Gprim(parent_unit).MakeInvisible()
        with cook.specialized_context(specialized_spawned, parent_unit):
            UsdGeom.Gprim(specialized_spawned).MakeInvisible()
        with cook.inherited_context(inherited_spawned):
            UsdGeom.Gprim(inherited_spawned).MakeInvisible()

        def _check_broadcasted_invisibility(asset, prim, method):
            target_stage = Usd.Stage.Open(asset)
            target_prefix = cook._broadcast_root_path(prim, method)
            authored = UsdGeom.Gprim(target_stage.GetPrimAtPath(target_prefix.AppendChild(Usd.ModelAPI(prim).GetAssetName()))).GetVisibilityAttr().Get()
            self.assertEqual(authored, 'invisible')

        for target_asset, target_prim, broadcast_type in (
                (parent_unit, parent_unit, Usd.Inherits),  # non-referenced, no context, asset unit is the target
                (inherited_spawned, inherited_spawned, Usd.Inherits),  # referenced asset unit, no context, asset unit is the target
                (parent_unit, specialized_spawned, Usd.Specializes),  # referenced, context unit is the target
        ):
            with self.subTest(target_asset=str(target_asset), target_prim=str(target_prim), broadcast_type=str(broadcast_type)):
                _check_broadcasted_invisibility(cook.unit_asset(target_asset), target_prim, broadcast_type)



from ._qt import QtCore, QtWidgets
from . import _graph

if __name__ == "__main__":
    QtCore.QCoreApplication.setAttribute(QtCore.Qt.AA_ShareOpenGLContexts)
    # python -m grill.views._graph
    app = QtWidgets.QApplication([])
    # TODO: create "plugs" from parsing label?
    nodes_info = {
        1: dict(
            label="{<one>x:y:z|<two>z}",
            style="rounded",  # these can be set at the graph level
            shape="record",
        ),
        2: dict(
            label="{<three>a|<four>b}",
            style='rounded',
            shape="record",
        ),
        3: dict(
            label="{<five>c|<six>d}",
            style='rounded',
            shape="record",
        ),
        "parent": dict(
            shape="box", fillcolor="#afd7ff", color="#1E90FF", style="filled,rounded"
        ),
        "child1": dict(
            shape="box", fillcolor="#afd7ff", color="#1E90FF", style="filled,rounded"
        ),
        "child2": dict(
            shape="box", fillcolor="#afd7ff", color="#1E90FF", style="invis"
        ),
    }
    edges_info = (
        (1, 1, dict(color='sienna:crimson:orange')),
        (1, 2, dict(color='crimson')),
        (2, 1, dict(color='seagreen')),
        (3, 2, dict(color='steelblue', tailport='five')),
        (3, 1, dict(color='hotpink', tailport='five')),
        ("parent", "child1"),
        ("parent", "child2", dict(label='invis')),
    )

    graph = _graph.nx.MultiDiGraph()
    graph.add_nodes_from(nodes_info.items())
    graph.add_edges_from(edges_info)
    graph.graph['graph'] = {'rankdir': 'LR'}
    graph.graph['edge'] = {"color": 'crimson'}
    outline_color = "#4682B4"  # 'steelblue'
    background_color = "#F0FFFF"  # 'azure'
    # graph.graph['node'] = {'shape': 'none', 'color': outline_color, 'fillcolor': background_color}
    table_row = '<tr><td port="{port}" border="0" bgcolor="{color}" style="ROUNDED">{text}</td></tr>'

    connection_nodes = dict(
        ancestor=dict(
            # fields=('ancestor', 'cycle_in', 'roughness', 'cycle_out', 'surface'),
            plugs={
                '': 0,
                'cycle_in': 1,
                'roughness': 2,
                'cycle_out': 3,
                'surface': 4
            },
            shape='none',
            connections=dict(
                surface=[('successor', 'surface')],
                cycle_out=[('ancestor', 'cycle_in')],
            ),
        ),
        successor=dict(
            plugs={'': 0, 'surface': 1},
            shape='none',
            connections=dict(),
        )
    )
    connection_edges = []

# A:\write\code\git\easy-edgedb\chapter10\assets\dracula-3d-Model-Place-rnd-main-GoldenKroneHotel-lead-base-whole.1.usda
    def _add_edges(src_node, src_name, tgt_node, tgt_name):
        tooltip = f"{src_node}.{src_name} -> {tgt_node}.{tgt_name}"
        print(tooltip)
        connection_edges.append(
            (src_node, tgt_node, {"tailport": src_name, "headport": tgt_name, "tooltip": tooltip}))


    for node, data in connection_nodes.items():
        label = f'<<table border="1" cellspacing="2" style="ROUNDED" bgcolor="{background_color}" color="{outline_color}">'
        label += table_row.format(port="", color="white",
                                  text=f'<font color="{outline_color}"><b>{node}</b></font>')
        # for index, plug in enumerate(data['plugs'], start=1):  # we start at 1 because index 0 is the node itself
        for plug, index in data['plugs'].items():  # we start at 1 because index 0 is the node itself
            if not plug:
                continue
            plug_name = plug
            sources = data['connections'].get(plug, [])  # (valid, invalid): we care only about valid sources (index 0)
            print(f"{node=}, {plug=}")
            print(sources)
            color = r"#F08080" if sources else background_color
            # color = plug_colors[type(plug)] if isinstance(plug, UsdShade.Output) or sources else background_color
            label += table_row.format(port=plug_name, color=color, text=f'<font color="#242828">{plug_name}</font>')
            for source_node, source_plug in sources:
                # node_id='ancestor', plug_name='cycle_out', ancestor, source.sourceName='cycle_in'
                # tooltip='/TexModel/boardMat/PBRShader.cycle_in -> /TexModel/boardMat/PBRShader.cycle_out'
                _add_edges(node, plug_name, source_node, source_plug)

        label += '</table>>'
        data['label'] = label
        data.pop('connections', None)

    graph.add_nodes_from(connection_nodes.items())
    graph.add_edges_from(connection_edges)

    widget = QtWidgets.QFrame()
    splitter = QtWidgets.QSplitter(QtCore.Qt.Horizontal)
    # return
    cached = {}
    from networkx import nx_pydot

    original = nx_pydot.write_dot


    # def _cached_dot(subgraph, fp):
    #     if not cached:
    #         print("WRITIGGGNNNNG")
    #         original(subgraph, fp)
    #         cached[True] = fp


    # cached_svg = {}
    # original_dot2svg = _graph._dot_2_svg


    # def _cached_dot2svg(fp):
    #     if not cached_svg:
    #         print(f"~~~~~~~~~~~~~~~~~~~ ~~~~~~~~~~~~~~~~~~~~ WITIHGN AND SAVING SVG: {fp}")
    #         result = original_dot2svg(fp)
    #         cached_svg[True] = result
    #         return result
    #     return cached_svg[True]


    from unittest import mock

    def _redirect_svg(self, filepath):
        return self._on_dot_result(r"A:\write\code\git\grill\tests\test_data\_mini_graph.svg")
        # return

    # with (
    #     # mock.patch(f"grill.views._graph.nx.nx_pydot.write_dot", new=_cached_dot),
    #     # mock.patch(f"grill.views._graph._dot_2_svg", new=_cached_dot2svg),
    #     mock.patch(f"grill.views._graph._DotViewer.setDotPath", new=_redirect_svg),
    #     # ...
    # ):
    from pxr import Usd, UsdShade, Sdf
    from . import description
    stage = Usd.Stage.CreateInMemory()
    material = UsdShade.Material.Define(stage, '/TexModel/boardMat')
    pbrShader = UsdShade.Shader.Define(stage, '/TexModel/boardMat/PBRShader')
    pbrShader.CreateInput("roughness", Sdf.ValueTypeNames.Float).Set(0.4)
    material.CreateSurfaceOutput().ConnectToSource(pbrShader.ConnectableAPI(), "surface")
    # Ensure cycles don't cause recursion
    cycle_input = pbrShader.CreateInput("cycle_in", Sdf.ValueTypeNames.Float)
    cycle_output = pbrShader.CreateOutput("cycle_out", Sdf.ValueTypeNames.Float)
    cycle_input.ConnectToSource(cycle_output)

    stage = Usd.Stage.Open(r"A:\write\code\git\easy-edgedb\chapter10\assets\dracula-3d-Model-Place-rnd-main-GoldenKroneHotel-lead-base-whole.1.usda")
    stage.GetPrimAtPath("/Origin").GetVariantSet("color").SetVariantSelection("random_uniform")

    for cls in _graph.GraphView, _graph._GraphSVGViewer:
            for pixmap_enabled in ((True, False) if cls == _graph._GraphSVGViewer else (False,)):
                _graph._GraphViewer = cls
                _graph._USE_SVG_VIEWPORT = pixmap_enabled

                child = description.LayerStackComposition(parent=widget)
                child.setStage(stage)
                # stack.show()

                # child = cls(parent=widget)
                # child._graph = graph
                # child.view(graph.nodes)
                # child.setMinimumWidth(150)
                splitter.addWidget(child)
                # viewer = description._ConnectableAPIViewer()
                # viewer.setPrim(material)
                # splitter.addWidget(viewer)

    layout = QtWidgets.QHBoxLayout()
    layout.addWidget(splitter)

    widget.setLayout(layout)
    widget.show()

    app.exec_()



import logging

from functools import cache, partial

from maya import cmds
from ._qt import QtWidgets

if cmds.about(qt=True).startswith("6"):
    from shiboken6 import wrapInstance
else:
    from shiboken2 import wrapInstance

import ufe
import mayaUsd
import maya.OpenMayaUI as omui
import maya.api.OpenMaya as om

from . import description as _description, sheets as _sheets, create as _create, _core, stats as _stats
_description._PALETTE.set(0)  # (0 == dark, 1 == light)

_logger = logging.getLogger(__name__)


@cache
def _main_window():
    return wrapInstance(int(omui.MQtUtil.mainWindow()), QtWidgets.QWidget)


def _stage_on_widget(widget_creator, _cache=True):
    def _launcher():
        widget = widget_creator(parent=_main_window())
        widget.setStyleSheet(
            # Maya checked buttons style look ugly (all black),
            # so re-use the USDView style for push buttons
            _core._USDVIEW_PUSH_BUTTON_STYLE + """
            /* Without this, maya shows larger, pixelated arrows when sorting tables. */
            QHeaderView::down-arrow { top: 1px; width: 13px; height:9px; subcontrol-position: top center;}
            """
        )
        usd_proxies = cmds.ls(typ='mayaUsdProxyShape', l=True)
        stage = next((mayaUsd.ufe.getStage(node) for node in usd_proxies), None)
        if stage:
            widget.setStage(stage)
        return widget
    if _cache:
        _launcher = cache(_launcher)
    return _launcher


@cache
def _prim_composition():
    widget = _description.PrimComposition(parent=_main_window())

    def selection_changed(*_, **__):
        for item in ufe.GlobalSelection.get():
            prim = mayaUsd.ufe.getPrimFromRawItem(item.getRawAddress())
            if prim.IsValid():
                widget.setPrim(prim)
                break
        else:
            widget.clear()

    om.MEventMessage.addEventCallback("UFESelectionChanged", selection_changed)
    selection_changed()
    return widget


@cache
def _connections_viewer():
    widget = _description._ConnectableAPIViewer(parent=_main_window())

    def selection_changed(*_, **__):
        for item in ufe.GlobalSelection.get():
            prim = mayaUsd.ufe.getPrimFromRawItem(item.getRawAddress())
            if prim.IsValid():
                widget.setPrim(prim)
                break
        else:
            widget.setPrim(None)

    om.MEventMessage.addEventCallback("UFESelectionChanged", selection_changed)
    selection_changed()
    return widget


@cache
def create_menu():
    _logger.info(f"Creating The Grill menu.")
    menu = cmds.menu("grill", label="👨‍🍳 Grill", tearOff=True, parent="MayaWindow")

    def show(_launcher, *_, **__):
        return _launcher().show()

    creation_menu_items = (
        ("Create Assets", _stage_on_widget(_create.CreateAssets)),
        ("Taxonomy Editor", _stage_on_widget(_create.TaxonomyEditor)),
    ) if _create.cook else ()

    for title, launcher in (
            *creation_menu_items,
            ("Spreadsheet Editor", _stage_on_widget(_sheets.SpreadsheetEditor)),
            ("Prim Composition", _prim_composition),
            ("Connections Viewer", _connections_viewer),
            ("LayerStack Composition", _stage_on_widget(_description.LayerStackComposition)),
            ("Stage Stats", _stage_on_widget(_stats.StageStats, _cache=False)),
    ):
        cmds.menuItem(title, command=partial(show, launcher), parent=menu)

    return menu



from ._qt import QtCore, QtWidgets
from pxr import Usd, Pcp, UsdGeom

def crash_edit_target(stage):
    "This crashes in usdview"
    prim = stage.GetPrimAtPath("/root/stoat/body_M_hrc/GEO/nose_M_geo")
    # index = prim.GetPrimIndex()
    index = prim.ComputeExpandedPrimIndex()
    root = index.rootNode

    def walk_composition(node):
        yield node
        for child in node.children:
            yield from walk_composition(child)

    target_node = next(
        (
            (index, node) for index, node in enumerate(walk_composition(root))
            if "mk020_0281_shotrange_edit_shot.usda" in node.layerStack.layerTree.layer.identifier
               and node.path == prim.GetPath()
               and node.arcType == Pcp.ArcTypeReference
        ),
        None
    )
    # first_inherits = next((child for child in root.children if child.arcType == Pcp.ArcTypeInherit), None)
    # assert first_inherits
    assert target_node
    target_node_index, target_node = target_node
    print(f"{target_node_index=}")
    edit_target = Usd.EditTarget(target_node.layerStack.layerTree.layer, target_node)
    with Usd.EditContext(stage, edit_target):
        UsdGeom.Gprim(prim).MakeInvisible()

    print(edit_target.GetLayer().ExportToString())


if __name__ == "__main__":
    QtCore.QCoreApplication.setAttribute(QtCore.Qt.AA_ShareOpenGLContexts)
    # python -m grill.views._graph
    app = QtWidgets.QApplication([])

    from . import description



    stage = Usd.Stage.Open(r"R:\dpel\ALab\ALab\entry.usda")
    crash_edit_target(stage)
    stage = Usd.Stage.Open(r"R:\dpel\ALab\ALab\entity\stoat01\stoat01.usda")
    # stage = Usd.Stage.Open(r"A:\write\code\git\easy-edgedb\chapter10\assets\dracula-3d-Model-Country-rnd-main-Romania-lead-base-whole.1.usda")
    prim = stage.GetPrimAtPath("/root/body_M_hrc/GEO/nose_M_geo")
    # prim = stage.GetPrimAtPath("/Origin/Buildings/GoldenKroneHotel/Geom/Basis_L/Basis_L_0")

    # index = prim.GetPrimIndex()
    # pattern = description._PCP_DUMP_PATTERN


    widget = QtWidgets.QFrame()
    splitter = QtWidgets.QSplitter(QtCore.Qt.Horizontal)
    child = description.PrimComposition(parent=widget)
    child.setPrim(prim)
    splitter.addWidget(child)
    layout = QtWidgets.QHBoxLayout()
    layout.addWidget(splitter)

    widget.setLayout(layout)
    widget.show()

    app.exec_()
